import os

from paste.deploy import loadapp
from paste.script.cherrypy_server import cpwsgi_server

if __name__ == "__main__":
    port = int(os.environ.get("PORT", 5000))
    wsgi_app = loadapp('config:production.ini', relative_to='.')
    cpwsgi_server(wsgi_app, host='0.0.0.0', port=port,
                  numthreads=10, request_queue_size=200)



from cheeseprism import index
from cheeseprism import utils
from cheeseprism.resources import App
from cheeseprism.utils import path
from contextlib import contextmanager
from mock import Mock
from mock import patch
from pyramid import testing
from pyramid.decorator import reify
from pyramid.events import subscriber
from pyramid.httpexceptions import HTTPFound
from stuf import stuf
from test_pipext import PipExtBase
import futures
import itertools
import unittest

here = path(__file__).parent


def test_instructions():
    from cheeseprism.views import instructions
    assert instructions(None, None)


class ViewTests(unittest.TestCase):

    def setUp(self):
        self.config = testing.setUp()

    def setup_event(self):
        self.event_results = {}
        from cheeseprism.event import IPackageEvent

        @subscriber(IPackageEvent)
        def test_event_fire(event):
            self.event_results.setdefault(event.__class__.__name__, []).append(event)

        self.config.add_subscriber(test_event_fire)

    def tearDown(self):
        self.event_results = None
        testing.tearDown()
        CPDummyRequest._namer = None
        if CPDummyRequest.test_dir is not None:
            CPDummyRequest.test_dir.rmtree()
            CPDummyRequest.test_dir = None
        CPDummyRequest._index_data = {}
        for d in CPDummyRequest.cleanup:
            d.rmtree_p()
        CPDummyRequest.cleanup = []

    @patch('cheeseprism.rpc.PyPi.search')
    def test_find_package(self, search_pypi):
        search_pypi.return_value = ['1.3a2']
        from cheeseprism.views import find_package
        request = testing.DummyRequest()
        request.POST['search_box'] = 'pyramid'
        out = find_package(App(request), request)
        assert out['releases'] == None
        assert out['search_term'] == None

        request.method = 'POST'
        out = find_package(App(request), request)
        assert out['releases'] == ['1.3a2'], "%s != %s" %(out['releases'], ['1.3a2'])
        assert out['search_term'] == 'pyramid'

    @patch('cheeseprism.rpc.PyPi.release_urls')
    def test_package_no_details(self, pd):
        """
        pypi doesn't know anything about our package
        """
        from cheeseprism.views import from_pypi
        pd.return_value = None
        request = testing.DummyRequest()
        request.matchdict.update(dict(name='boto',
                                      version='1.2.3'))
        out = from_pypi(request)
        assert isinstance(out, HTTPFound)
        assert out.location == '/find-packages'

    @patch('cheeseprism.rpc.PyPi.release_urls')
    def test_package_md5_matches(self, pd):
        """
        package is already in index
        """
        from cheeseprism.views import from_pypi
        td = dict(name='boto',
                  version='1.2.3',
                  md5_digest='12345',
                  filename='boto-1.2.3.tar.gz')
        pd.return_value = [td]
        request = CPDummyRequest()
        request.matchdict.update(td)
        request.index_data.update({'12345':True})
        out = from_pypi(request)
        assert isinstance(out, HTTPFound)
        assert out.location == '/index/boto'

    @patch('cheeseprism.rpc.PyPi.release_urls')
    def test_package_httperror(self, pd):
        """
        from_pypi: test catching httperror
        """
        request = self.package_request(pd)
        with patch('requests.get') as get:
            from cheeseprism import views; reload(views)
            get.side_effect = views.HTTPError('http://boto', 500, 'KABOOM', dict(), None)
            out = views.package(request)
        assert isinstance(out, HTTPFound)

    @patch('cheeseprism.rpc.PyPi.release_urls')
    def test_package_urlerror(self, pd):
        """
        from_pypi: test catching urlerror
        """
        request = self.package_request(pd)
        with patch('requests.get') as get:
            from cheeseprism import views; reload(views)
            get.side_effect = views.URLError('kaboom')
            out = views.package(request)
        assert isinstance(out, HTTPFound)
        assert out.location == '/find-packages'

    @patch('cheeseprism.rpc.PyPi.release_urls')
    def test_package_good(self, pd):
        """
        from_pypi: test catching urlerror
        """
        request = self.package_request(pd)
        with patch('requests.get') as get:
            resp = get.return_value = Mock('response')
            resp.content = PipExtBase.dists['dp'].bytes()
            from cheeseprism import views; reload(views)
            out = views.package(request)
        assert isinstance(out, HTTPFound)
        assert out.location == '/index/boto'

    @patch('cheeseprism.rpc.PyPi.release_urls')
    def test_package_downloads_ok_but_bad(self, pd):
        """
        from_pypi: test catching urlerror
        """
        request = self.package_request(pd)
        with patch('requests.get') as get:
            resp = get.return_value = Mock('response')
            resp.content = PipExtBase.dists['dp'].bytes()
            from cheeseprism import views; reload(views)
            with patch('cheeseprism.index.IndexManager.pkginfo_from_file') as pkff:
                pkff.side_effect = ValueError("KABOOM")
                out = views.package(request)
        assert isinstance(out, HTTPFound)
        assert out.location == '/find-packages'

    def package_request(self, pd, td=None):
        if td is None:
            td = dict(name='boto',
                      version='1.2.3',
                      md5_digest='12345',
                      url='http://boto',
                      filename='boto-1.2.3.tar.gz')
        pd.return_value = [td]
        request = CPDummyRequest()
        request.matchdict.update(td)
        return request

    def test_index_view(self):
        from cheeseprism.views import upload as index
        request = testing.DummyRequest()
        assert index(App(request), request) == {}

    @property
    def base_cr(self):
        self.curreq = request = CPDummyRequest()
        return App(request), request

    def test_regenerate_index(self):
        """
        Basic regeneration of entire index
        """
        from cheeseprism.views import regenerate_index
        context, request = self.base_cr
        out = regenerate_index(context, request)
        assert out['disabled'] == False
        assert isinstance(out, dict)

        context, req = self.base_cr
        req.method = "POST"

        out = regenerate_index(context, req)
        assert isinstance(out, HTTPFound)
        assert req.file_root.exists(), "index not created at %s" %req.file_root
        assert len(req.file_root.files()) == 2
        assert set([x.name for x in req.file_root.files()]) == set(('index.html', 'index.json'))

    def test_upload_raises(self):
        from cheeseprism.views import upload
        context, request = self.base_cr
        request.POST['content'] = ''
        request.method = 'POST'
        with self.assertRaises(RuntimeError):
            upload(context, request)

    def test_upload_raises_packageadded(self):
        """
        If adding the package raises an error, an exception should be logged
        """
        from cheeseprism.views import upload
        self.setup_event()
        context, request = self.base_cr
        request.method = 'POST'
        request.POST['content'] = FakeFS(path('dummypackage2/dist/dummypackage-0.1.tar.gz'))
        with patch('cheeseprism.views.event.PackageAdded',
                   side_effect=RuntimeError('Kaboom')):
            with patch('path.path.write_bytes'):
                with self.assertRaises(RuntimeError):
                    upload(context, request)

    @patch('path.path.write_bytes')
    def test_upload(self, wb):
        from cheeseprism.views import upload
        self.setup_event()
        context, request = self.base_cr
        request.method = 'POST'
        request.POST['content'] = FakeFS(path('dummypackage/dist/dummypackage-0.0dev.tar.gz'))
        with patch('cheeseprism.index.IndexManager.pkginfo_from_file',
                   return_value=stuf(name='dummycode', version='0.0dev')) as pkif:
            res = upload(context, request)
            assert pkif.called
            assert 'PackageAdded' in self.event_results
            assert self.event_results['PackageAdded'][0].name == pkif.return_value.name
        assert res.headers == {'X-Swalow-Status': 'SUCCESS'}

    def test_upload_w_rename(self):
        from cheeseprism.views import upload
        self.setup_event()
        context, request = self.base_cr

        request._namer = utils.strip_master
        request.method = 'POST'
        request.POST['content'] = FakeFS(path('mastertest-0.0-master.tar.gz'))

        def test_pkif(p, moe):
            assert p.basename() == u'mastertest-0.0.tar.gz'
            return stuf(name='mastertest', version='0.0-master')

        with patch('cheeseprism.index.IndexManager.pkginfo_from_file',
                   side_effect=test_pkif) as pkif:

            res = upload(context, request)
            assert pkif.called

        assert res.headers == {'X-Swalow-Status': 'SUCCESS'}

    def test_from_requirements_GET(self):
        from cheeseprism.views import from_requirements
        context, request = self.base_cr
        assert from_requirements(context, request) == {}

    def test_from_requirements_POST(self):
        from cheeseprism.views import from_requirements
        context, request = self.base_cr
        flash = request.session.flash = Mock('flash')
        notify = request.registry.notify = Mock('notify')
        request.method = 'POST'
        reqfile = here / 'req-1.txt'
        request.POST['req_file'] = FakeFS(reqfile, body=reqfile.text())
        with mock_downloader():
            out = from_requirements(context, request)
            assert out
            assert flash.called
            assert notify.called


@contextmanager
def mock_downloader():
    with patch('cheeseprism.pipext.RequirementDownloader') as dl:
        dler = dl.return_value = Mock(name='downloader')
        pkgi, outf = Mock(name='pkginfo'), Mock(name='outfile')
        outf.filename = 'outfile'
        pkgi.name = "dummyfile"
        dler.download_all.return_value = (pkgi, outf),
        dler.skip = (outf,)
        dler.errors = ('error',)
        dl.req_set_from_file.return_value = (dl, Mock(name='finder'))
        yield dl


class DummyResponse(object):
    def __init__(self):
        self.headers = {}


class FakeFS(object):
    def __init__(self, path, body="Some gzip binary"):
        self.filename = path.name
        self.file = Mock()
        self.file.read.return_value = body


class CPDummyRequest(testing.DummyRequest):
    test_dir = None
    counter = itertools.count()
    env = None
    _index_data = {}
    _namer = None
    cleanup = []
    disable_regen = False

    @reify
    def namer(self):
        if self._namer is None:
            from cheeseprism.utils import secure_filename
            return secure_filename
        return self._namer

    @property
    def userid(self):
        return 'bob'

    @property
    def settings(self):
        return {}

    @property
    def file_root(self):
        test_dir = getattr(self, 'test_dir')
        if test_dir is None:
            test_dir = '%s-view-tests' %next(self.counter)
            self.test_dir = test_dir = path(__file__).parent / "test-indexes" / test_dir
            self.test_dir.mkdir_p()
            self.cleanup.append(test_dir)
        return self.test_dir

    @property
    def index_templates(self):
        if self.env is None:
            self.env = index.EnvFactory.from_str()
        return self.env

    @property
    def index(self):
        return index.IndexManager(self.file_root,
                                  template_env=self.index_templates,
                                  executor=futures.ThreadPoolExecutor(max_workers=4))

    @reify
    def response(self):
        return DummyResponse()

    @reify
    def index_data_path(self):
        return self.file_root / 'index.json'

    @reify
    def index_data(self):
        return self._index_data



from pyramid.decorator import reify
from path import path as path_base
import functools
import hashlib
import logging
import os
import pkg_resources
import re
import time


logger = logging.getLogger(__name__)


class path(path_base):

    @reify
    def md5hex(self):
        return self.read_md5_fast().encode('hex')

    def read_md5_fast(self):
        """ Calculate the md5 hash for this file.

        This reads through the entire file.
        """
        return self.read_hash_fast('md5')

    def _hash_whole(self, hash_name):
        """ Returns a hash object for the file at the current path.

            `hash_name` should be a hash algo name such as 'md5' or 'sha1'
            that's available in the `hashlib` module.
        """
        m = hashlib.new(hash_name)
        m.update(self.bytes())
        return m

    def read_hash_fast(self, hash_name):
        """ Calculate given hash for this file.

        List of supported hashes can be obtained from hashlib package. This
        reads the entire file.
        """
        return self._hash_whole(hash_name).digest()



def resource_spec(spec):
    """
    Loads resource from a string specifier.
    >>> from doula.utils import resource_spec
    >>> resource_spec('egg:ReleaseDoula#doula/roles.yml')
    '.../doula/roles.yml'

    >>> resource_spec('file:data/languages.ini')
    'data/languages.ini'

    >>> resource_spec('data/languages.ini')
    'data/languages.ini'
    """
    filepath = spec
    if spec.startswith('egg:'):
        req, subpath = spec.split('egg:')[1].split('#')
        req = pkg_resources.get_distribution(req).as_requirement()
        filepath = pkg_resources.resource_filename(req, subpath)
    elif spec.startswith('file:'):
        filepath = spec.split('file:')[1]
    # Other specs could be added, but egg and file should be fine for
    # now
    return filepath


_windows_device_files = ('CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1',
                         'LPT2', 'LPT3', 'PRN', 'NUL')
_filename_ascii_strip_re = re.compile(r'[^A-Za-z0-9_.-]')


def secure_filename(filename):
    r"""Pass it a filename and it will return a secure version of it.  This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.  The filename returned is an ASCII only string
    for maximum portability.

    On windows system the function also makes sure that the file is not
    named after one of the special device files.

    >>> secure_filename("My cool movie.mov")
    'My_cool_movie.mov'
    >>> secure_filename("../../../etc/passwd")
    'etc_passwd'
    >>> secure_filename(u'i contain cool \xfcml\xe4uts.txt')
    'i_contain_cool_umlauts.txt'

    The function might return an empty filename.  It's your responsibility
    to ensure that the filename is unique and that you generate random
    filename if the function returned an empty one.

    .. versionadded:: 0.5

    :param filename: the filename to secure

    Credit: from Armin Ronacher's Werkzeug, BSD
    """
    if isinstance(filename, unicode):
        from unicodedata import normalize
        filename = normalize('NFKD', filename).encode('ascii', 'ignore')
    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')
    filename = str(_filename_ascii_strip_re.sub('', '_'.join(
                   filename.split()))).strip('._')

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    if os.name == 'nt' and filename and \
       filename.split('.')[0].upper() in _windows_device_files:
        filename = '_' + filename

    return filename


def strip_master(filename):
    """
    Create a secure filename and remove the string '-master'
    """
    filename = secure_filename(filename)
    return filename.replace('-master', '')


class benchmark(object):
    """
    from DaBeaz http://bit.ly/15nNrlF
    """
    def __init__(self, name, logger=logger.debug):
        self.name = name
        self.logger = logger

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, ty, val, tb):
        end = time.time()
        self.logger("%s: %0.3f seconds", self.name, end-self.start)
        return False

    def __call__(self, func):
        @functools.wraps(func)
        def wrapped(*args, **kw):
            with self:
                return func(*args, **kw)
        return wrapped



from .jenv import EnvFactory
from cheeseprism.auth import BasicAuthenticationPolicy
from cheeseprism.resources import App
from pyramid.config import Configurator
from pyramid.session import UnencryptedCookieSessionFactoryConfig
from pyramid.settings import asbool
import futures
import logging
import os


logger = logging.getLogger(__name__)


def main(global_config, **settings):
    settings = dict(global_config, **settings)
    settings.setdefault('jinja2.i18n.domain', 'CheesePrism')

    session_factory = UnencryptedCookieSessionFactoryConfig('cheeseprism')
    config = Configurator(root_factory=App, settings=settings,
                          session_factory=session_factory,
                          authentication_policy=\
                          BasicAuthenticationPolicy(BasicAuthenticationPolicy.noop_check))

    setup_workers(config.registry)

    config.add_translation_dirs('locale/')

    config.include('.request')
    config.include('.views')
    config.include('.index')

    tempspec = settings.get('cheeseprism.index_templates', '')
    config.registry['cp.index_templates'] = EnvFactory.from_str(tempspec)

    if asbool(settings.get('cheeseprism.pipcache_mirror', False)):
        config.include('.sync.pip')

    if asbool(settings.get('cheeseprism.auto_sync', False)):
        config.include('.sync.auto')

    tempfile_limit = int(settings.get('cheeseprism.temp_file_limit', 10*1024))
    config.add_request_method(lambda req: tempfile_limit,
                              name='request_body_tempfile_limit', reify=True)

    config.add_request_method(lambda req: asbool(settings.get('cheeseprism.disable.regenerate', False)),
                              name='disable_regen', reify=True)

    return config.make_wsgi_app()


def ping_proc(i):
    pid = os.getpid()
    logger.debug("worker %s up: %s", i, pid)
    return pid


def setup_workers(registry):
    """
    now thread only
    """
    settings = registry.settings

    registry['cp.executor_type'] = 'thread'

    executor = futures.ThreadPoolExecutor

    workers = int(settings.get('cheeseprism.futures.workers', 5))


    logging.info("Starting thread executor w/ %s workers", workers)
    executor = registry['cp.executor'] = executor(workers)

    # -- This initializes our threads
    workers = [str(pid) for pid in executor.map(ping_proc, range(workers))]
    logger.info("workers: %s", " ".join(workers))



from zope.interface import Attribute
from zope.interface import Interface
from zope.interface import implements

class IIndexEvent(Interface):
    """
    An lower level event involving the index
    """

class IIndexUpdate(Interface):
    """
    An low level event involving the index
    """

class IPackageEvent(IIndexEvent):
    """
    An event involving a package
    """
    path = Attribute('Path to package')


class IPackageAdded(IPackageEvent):
    """
    A package is added to the repository
    """


class IPackageRemoved(IPackageEvent):
    """
    A package is removed to the repository
    """


class IndexEvent(object):
    implements(IIndexEvent)
    def __init__(self, datafile, index, pkgdatas=None):
        self.index = index
        self.datafile = datafile
        self.pkgdatas = pkgdatas


class IndexUpdate(IndexEvent):
    implements(IIndexUpdate)


class PackageEvent(object):
    """
    Baseclass for pacakage events
    """
    implements(IPackageEvent)

    def __init__(self, index_manager, path=None, name=None, version=None):
        self.name = name
        self.version = version
        self.im = index_manager
        self.path = path

        if self.name is None and self.path:
            info = self.im.pkginfo_from_file(path, self.im.move_on_error)
            self.name = info.name
            self.version = info.version

            
class PackageAdded(PackageEvent):
    implements(IPackageAdded)
    
    def __init__(self, index_manager, path=None, name=None, version=None, rebuild_leaf=True):
        super(PackageAdded, self).__init__(index_manager, path, name, version)
        self.rebuild_leaf = rebuild_leaf


class PackageRemoved(PackageEvent):
    implements(IPackageRemoved)



"""
CheesePrism
===========

 - wsgiapp: main wsgi entry point

 - request: custom request object

 - pipext: utilities for parsing and downloading via requirements
           files

 - index: main code for managing file based python index, includes
          package management event subscribers

 - resources: pyramid in memory resource tree

 - event: package management events

 - utils: helpers, etc

 - desc: descriptors

"""


    




class updict(dict):
    """
    A descriptor that updates it's internal represention on set, and
    returns the dictionary to original state on deletion.
    """
    
    def __init__(self, *args, **kw):
        super(updict, self).__init__(*args, **kw)
        self.default = self.copy()

    def __get__(self, obj, objtype):
        return self

    def __set__(self, obj, val):
        self.update(val)

    def __delete__(self, obj):
        self.clear()
        self.update(self.default)


class template(object):
    """
    A little descriptor for returning templates from jinja env.
    """
    env_method = 'template_env'
    def __init__(self, name):
        self.name = name

    def get_env(self, obj):
        return getattr(obj, self.env_method, None)

    def __get__(self, obj, objtype):
        env = self.get_env(obj)
        return env.get_template(self.name)



from .utils import path
import argparse
import sys


def mc(args=sys.argv[1:]):
    def_save = path('old-master/files')
    parser = argparse.ArgumentParser()
    parser.add_argument("path", help="Path to index folder", type=path)
    parser.add_argument("--copydir", help="dir to co", type=path, action='store', default=def_save)

    pargs = parser.parse_args(args)
    index = pargs.path.abspath()
    copydir = pargs.copydir
    if copydir == def_save:
        copydir = index / copydir

    copydir.makedirs_p()

    files = index.files('*-master.tar.gz')

    print "%s files" %len(files)
    for fp in index.files('*-master.tar.gz'):
        np = fp.parent / fp.replace('-master', '')
        if np.exists():
            print "DUPE: %s" %np
        else:
            newfp = fp.rename(copydir / fp.basename())
            newfp.symlink(index / newfp.basename())
            newfp.copy(np)



from cheeseprism.utils import path
from cheeseprism.utils import resource_spec
from itertools import count
from mock import Mock
from mock import patch
from pip.exceptions import DistributionNotFound
from pip.index import Link
from pip.index import PackageFinder
from pprint import pformat as pp
from urllib2 import HTTPError
import logging
import pkginfo
import unittest

logger = logging.getLogger(__name__)
here = path(__file__).parent


class PipExtBase(unittest.TestCase):
    counter = count()
    index_parent = "egg:CheesePrism#tests/test-indexes"
    dists = dict(dp=(here / 'dummypackage/dist/dummypackage-0.0dev.tar.gz'),
                 dp2=(here / 'dummypackage2/dist/dummypackage-0.1.tar.gz'),
                 se=(here / 'something_else/dist/something_else-2.0.tar.gz'))

    filename = here / 'req-1.txt'

    @classmethod
    def get_base(cls):
        return path(resource_spec(cls.index_parent))

    @property
    def base(self): return self.get_base()

    def makeone(self):
        from cheeseprism import pipext
        rs, finder = self.make_finder_and_reqset()
        return pipext.RequirementDownloader(rs, finder=finder)

    def make_finder_and_reqset(self):
        from cheeseprism import pipext
        return pipext.RequirementDownloader.req_set_from_file(self.filename, self.download_dir)

    def setUp(self):
        self.count = next(self.counter)
        self.download_dir = self.base / ("%s-%s" %(self.count, 'req-dl'))
        if not self.download_dir.exists():
            self.download_dir.makedirs()

    def tearDown(self):
        logger.debug("teardown: %s", self.count)
        if self.base.exists():
            dirs = self.base.dirs()
            logger.info(pp(dirs))
            logger.info(pp([x.rmtree() for x in dirs]))


class TestReqDownloader(PipExtBase):
    """
    Catch all coverage tests for RequirementsDownloader
    """

    def test_init(self):
        """
        Sanity test creation
        """
        rd = self.makeone()
        reqs = rd.req_set.requirements
        assert len(reqs.values()) == 1
        assert reqs.keys() == ['dummypackage'], reqs.keys()
        self.download_dir = ''
        rd = self.makeone()

    @patch('requests.get')
    def test_download_url(self, mockget):
        mockget.return_value = Mock()
        mockget.return_value.content = self.dists['dp'].bytes()
        rd = self.makeone()
        pinfo, outfile = rd.download_url(Link('http://fakeurl.com/dummypackage.tar.gz#egg=dummypackage'))
        assert outfile.exists()
        assert pinfo.name == 'dummypackage'

    def test_readzip(self):
        rd = self.makeone()
        archive, name = Mock(), 'blah'
        out = rd.readzip(archive, name)
        assert out
        method, args, _ = archive.method_calls[0]
        assert method == 'read'
        assert args == (name,)

    @patch('zipfile.ZipFile')
    def test_depinfo_for_zip(self, zfmock):
        from cheeseprism import pipext
        zfmock.return_value = ZFMock()
        assert pipext.RequirementDownloader.depinfo_for_file('fake.zip') == ([], [])

    def test_package_finder_cm(self):
        from cheeseprism import pipext
        pf = pipext.RequirementDownloader.package_finder
        finder = pf(deplinks=['a deplink'], index_urls=['an index_url'])
        assert finder.dependency_links == ['a deplink']
        assert finder.index_urls == ['http://pypi.python.org/simple', 'an index_url']


class ZFMock(Mock):
    namelist = lambda self: []
    read = lambda self, name : ""


@patch('cheeseprism.pipext.RequirementDownloader.download_url')
class TestReqDownloaderHandler(PipExtBase):

    @property
    def mock_finder(self):
        self.link = Link('http://pkgurl/pkg.tar.gz#md5=12345')
        finder = Mock(spec_set=PackageFinder)
        finder.find_requirement.return_value = self.link
        return finder

    def raise_http_error(self, url, *args, **kw):
        raise HTTPError(url, 500, 'kaboom: %s %s' %(pp(args), pp(kw)), dict(), None)

    def get_pkginfo(self, dist):
        pkg = self.dists[dist]
        return pkginfo.sdist.SDist(pkg)

    def test_editable_req_error(self, download_url):
        rd = self.makeone()
        req = rd.req_set.requirements.values().pop()
        req.editable = True
        assert rd.handle_requirement(req, self.mock_finder) is None
        assert rd.errors
        assert 'Editable' in rd.errors[0]

    def raise_distnotfound(self, *args, **kw):
        raise DistributionNotFound("not found")

    def test_distribution_notfound_error(self, download_url):
        finder = self.mock_finder
        rd = self.makeone()
        finder.find_requirement.side_effect = self.raise_distnotfound
        req = rd.req_set.requirements.values().pop()
        download_url.return_value = (self.get_pkginfo('dp'), self.dists['dp']) # for failing test
        assert rd.handle_requirement(req, finder) is None

    def basic_prep(self, download_url):
        rd = self.makeone()
        req = rd.req_set.requirements.values().pop()
        download_url.return_value = (self.get_pkginfo('dp'), self.dists['dp'])
        return rd, req

    def test_skip(self, download_url):
        rd, req = self.basic_prep(download_url)
        finder = self.mock_finder
        rd.seen.add(self.link.hash)
        assert rd.handle_requirement(req, finder) is None

    def test_typeerror_raise(self, download_url):
        """
        Testing formality to make sure our diaper does not hide our
        mocks.
        """
        rd, req = self.basic_prep(download_url)
        download_url.side_effect = TypeError('Kaboom')
        with self.assertRaises(TypeError):
            assert rd.handle_requirement(req, self.mock_finder) is None

    def test_general_download_exception(self, download_url):
        """
        Check the diaper records what is dumped into it.

        At this point in `handle_requirements`, weird stuff happens
        with peoples distributions.  We can't save the world, but we
        can catch their crap and let the rest of the downloads
        continue.
        """
        rd, req = self.basic_prep(download_url)
        download_url.side_effect = Exception('Kaboom')
        assert rd.handle_requirement(req, self.mock_finder) is None
        assert len(rd.errors) == 1
        assert 'archive' in rd.errors[0]


    def test_handle_requirement_httperror(self, download_url):
        rd = self.makeone()
        req = rd.req_set.requirements.values().pop()
        download_url.side_effect = self.raise_http_error
        assert rd.handle_requirement(req, self.mock_finder) is None
        assert len(rd.errors) == 1
        assert 'download' in rd.errors[0]

    def test_handle_requirement_noreqs(self, download_url):
        rd = self.makeone()
        req = rd.req_set.requirements.values().pop()
        download_url.return_value = (self.get_pkginfo('dp'), self.dists['dp'])
        (pkginfo, path_to_sdist, deps) = rd.handle_requirement(req, self.mock_finder)
        assert deps is None, deps
        assert pkginfo.name == 'dummypackage'
        assert path_to_sdist == self.dists['dp']

    def test_handle_requirement_w_reqs(self, download_url):
        rd = self.makeone()
        req = rd.req_set.requirements.values().pop()
        download_url.return_value = (self.get_pkginfo('dp2'), self.dists['dp2'])
        (pkginfo, path_to_sdist, deps) = rd.handle_requirement(req, self.mock_finder)
        assert pkginfo.name == 'dummypackage'
        assert path_to_sdist == self.dists['dp2']
        assert deps.requirements.keys() == ['something-else'], deps.requirements.keys()


@patch('cheeseprism.pipext.RequirementDownloader.handle_requirement')
class TestReqDownloaderAll(PipExtBase):

    def test_download_all_nothing(self, handle):
        handle.return_value = None
        rd = self.makeone()
        with self.assertRaises(StopIteration):
            next(rd.download_all())

    def test_download_single_pkg(self, handle):
        marker = object()
        handle.return_value = marker, marker, None,
        rd = self.makeone()
        output = [x for x in rd.download_all()]
        assert len(output) == 1
        assert output[0] == (marker, marker)

    def test_download_w_deps(self, handle):
        marker1 = object()
        handle.return_value = marker1, marker1, RSMock,
        rd = self.makeone()

        out_gen = rd.download_all()
        assert next(out_gen) == (marker1, marker1)

        marker2 = object()
        handle.return_value = marker2, marker2, None,
        assert next(out_gen) == (marker2, marker2)

class RSMock(Mock):
    requirements = dict(something_else="something_else_requirement")



from ConfigParser import ConfigParser
from cheeseprism.utils import path
from cheeseprism.utils import resource_spec
from itertools import count
from mock import patch
from pprint import pformat as pprint
from pyramid.decorator import reify
import logging
import time
import unittest


logger = logging.getLogger(__name__)
here = path(__file__).parent


class FunctionalTests(unittest.TestCase):
    testdir = here / 'test-indexes'
    dummy = here / "dummypackage/dist/dummypackage-0.0dev.tar.gz"
    counter = count()
    index_parent = here / "test-indexes"
    pipcache = here / "pipcache"
    devini = here / 'development.ini'

    dummy = here / "dummypackage/dist/dummypackage-0.0dev.tar.gz"

    @classmethod
    def get_base(cls):
        return path(resource_spec(cls.index_parent))

    base = reify(lambda self: self.get_base())

    def setUp(self):
        self.count = next(self.counter)
        self.dummy.copy(self.testdir)
        self.dummypath = self.testdir / self.dummy.name

    def makeone(self, xtra=None, index_name='test-func-index', count=None):
        from cheeseprism.wsgiapp import main
        cp = ConfigParser(dict(here=self.base))

        with open(resource_spec(self.devini)) as fp:
            cp.readfp(fp)

        defaults = dict((x, cp.get('DEFAULT', x)) for x in cp.defaults())

        count = count is None and self.count or count
        self.idxpath = index_path = self.base / ("%s-%s" %(count, index_name))

        settings = {
            'cheeseprism.file_root': index_path,
            'cheeseprism.data_json': 'data.json'
            }

        settings = xtra and dict(settings, **xtra) or settings
        app = main(defaults, **settings)

        from webtest import TestApp
        return TestApp(app)

    def test_async_bulkupdate(self):
        idxname = 'async_bulk_up'
        idxpath = self.base / "0-" + idxname
        idxpath.mkdir_p()

        self.dummy.copy(idxpath)

        testapp = self.makeone({'cheeseprism.async_restart':'true'},
                               index_name=idxname, count=0)
        time.sleep(0.02)
        res = testapp.get('/index', status=200)
        assert 'dummy' in res.body

    def test_root_thead_pip_sync(self):
        with patch.dict('os.environ', {'PIP_DOWNLOAD_CACHE': resource_spec(self.pipcache)}):
            testapp = self.makeone({'cheeseprism.futures':'thread',
                                    'cheeseprism.pipcache_mirror':'true'})
            res = testapp.get('/index', status=200)
        assert 'WUT' in res.body

    def test_root_thread(self):
        testapp = self.makeone()
        res = testapp.get('/', status=200)
        self.failUnless('Cheese' in res.body)

    def test_request_size_limit(self):
        testapp = self.makeone({'cheeseprism.temp_file_limit': 100*1024,
                                'pyramid.includes': __name__ + '.request_size_check'})

        res = testapp.get('/sizetest', status=200)
        assert res.json.get('request_size', False) == 102400

    def tearDown(self):
        logger.debug("teardown: %s", self.count)
        if self.base.exists():
            dirs = self.base.dirs()
            logger.debug(pprint(dirs))
            time.sleep(0.02)
            logger.debug(pprint([x.rmtree() for x in dirs]))


def request_size_check(config):
    def sizetest(request):
        return {'request_size': request.request_body_tempfile_limit}
    config.add_route('sizetest', '/sizetest')
    config.add_view(sizetest, route_name='sizetest', renderer='json')



from . import utils
from .index import IndexManager
from .index import bulk_add_pkgs
from .utils import path
from threading import Thread
import logging
import os
import time

logger = logging.getLogger(__name__)


def sync_folder(index, folder):
    with utils.benchmark("Sync packages"):
        EXTS = index.archive_tool.EXTS
        candidates = dict((x.md5hex, x)\
                          for x in folder.files()\
                          if EXTS.match(x))

        current = dict((x.md5hex, x)\
                       for x in index.files)

        new = set(candidates) - set(current)
        for md5 in new:
            cpath = candidates.get(md5, None)
            if cpath is None:
                logger.error("missing md5: %s", md5)
                break
            _, name = cpath.rsplit('%2F', 1)
            cpath.copy(index.path / name)


def update_index(index, reg):
    with utils.benchmark('Update index after sync'):
        new_pkgs = index.update_data()
        if index.path.exists(): #for testing
            with utils.benchmark("New packages notifier: %s" %new_pkgs):
                leaves, archs = bulk_add_pkgs(index, new_pkgs, register=False)

            if index.write_index_html is True \
              and not all((index.datafile_path.exists(), index.home_file.exists())):
                index.write_index_home(index.projects_from_archives())

            return leaves, archs


def wait_for_file(fp, sleep=0.2, max_tries=5):
    fp = path(fp)
    if not fp.exists():
        logger.info('Waiting for base dir: %s' %fp)
        for x in range(max_tries):
            if not fp.exists():
                time.sleep(0.5)
                logger.info('bueller?')
            else:
                logger.info('Found based dir: %s' %fp)
                yield fp
    raise RuntimeError('%s not found in %ss' %(fp, max_tries * sleep))


def sync_cache(index, registry):
    pdc = path(os.environ.get('PIP_DOWNLOAD_CACHE'))
    if not pdc.exists():
        logger.error("Environmental var $PIP_DOWNLOAD_CACHE not set or %s not yet created" %pdc)
        return pdc

    wait_for_file(index.path)
    wait_for_file(index.datafile_path)
    sync_folder(index, pdc)
    try:
        update_index(index, registry)
    except :
        logger.exception("sync_cache:update_index failed")



def pip(config):
    index = IndexManager.from_registry(config.registry)
    thread = Thread(target=sync_cache, args=(index, config.registry), name='pip-updater')
    thread.start()


def dowatch(index, reg, pdc):
    dfexists = index.datafile_path.exists()

    assert dfexists
    data = index.data_from_path(index.datafile_path)

    index_reg = set()
    for md5, info in data:
        arch = path(info['filename'])
        assert arch.exists(), "index.json: %s missing" %arch
        index_reg.add((arch, arch.stat()))

    fsreg = set()
    for arch in index.files:
        # this assertion should rarely fail unless someone is sleeping out
        assert arch.exists(), "fs: %s missing" %arch
        fsreg.add((arch, arch.stat()))

    assert fsreg == index_reg, "index.json and filesystem do not match: %s" % index_reg ^ fsreg


def index_watch(index, reg, interval=3, failint=3, pdc=None, dowatch=dowatch):
    time.sleep(10)
    while True:
        try:
            dowatch(index, reg, pdc)
        except AssertionError:
            logger.exception("Index fails consistency tests")
            index.regenerate_all()
        except KeyboardInterrupt:
            raise
        except Exception:
            logger.exception("Who watches the watchman's exceptions?")
            time.sleep(3)
        finally:
            time.sleep(interval)

from pyramid.path import DottedNameResolver as dnr
resolve = dnr(None).maybe_resolve


def auto(config):
    dowatch = resolve(config.registry.settings.get('cheeseprism.dowatch', 'cheeseprism.sync.dowatch'))
    index = IndexManager.from_registry(config.registry)
    thread = Thread(target=index_watch, args=(index, config.registry, path(os.environ['PIP_DOWNLOAD_CACHE'])), kwargs=dict(dowatch=dowatch))
    thread.start()
    #@@ configure additional folders?



def test_strip_master():
    from cheeseprism.utils import strip_master
    assert strip_master('wat-10.1-master.tar.gz') == 'wat-10.1.tar.gz'
    assert strip_master('wat-10.1.tar.gz') == 'wat-10.1.tar.gz' 



#from pyramid.url import resource_url

class BaseResource(dict):
    """
    Base class for resources
    """

##     @property
##     def approot(self):
##         return self.recurse_parent(App)
            
##     def recurse_parents(self, klass, parent=None):
##         if parent is None:
##             parent = self.__parent__
##         else:
##             parent = parent.__parent__
##         if parent is None:
##             # we've rooted out
##             return None
##         if isinstance(parent, klass):
##             return parent
##         return self.recurse_parent(klass, parent)
    
##     def suburls(self, request):
##         for name in sorted(self):
##             obj = self[name]
##             url = resource_url(self, request, name)
##             yield obj, url


class App(BaseResource):
    __name__ = ''
    __parent__ = None
    
    def __init__(self, request):
        self.request = request



from cheeseprism.utils import path
from cheeseprism.utils import resource_spec
from itertools import count
from mock import Mock
from mock import patch
from pprint import pformat as pprint
from stuf import stuf
import futures
import logging
import textwrap
import unittest

logger = logging.getLogger(__name__)
here = path(__file__).parent


def test_data_from_path():
    from cheeseprism import index
    datafile = here / 'index.json'
    assert index.IndexManager.data_from_path(datafile) == {}
    datafile.write_text("{}")
    assert index.IndexManager.data_from_path(datafile) == {}


class IndexTestCase(unittest.TestCase):
    counter = count()
    index_parent = here / "test-indexes"

    dummy = here / "dummypackage/dist/dummypackage-0.0dev.tar.gz"
    dum_whl = here / "dummypackage/dist/dummypackage-0.0dev-py27-none-any.whl"

    @classmethod
    def get_base(cls):
        return path(resource_spec(cls.index_parent))

    @property
    def base(self): return self.get_base()

    def make_one(self, index_name='test-index', pkg='dummy'):
        from cheeseprism import index
        self.count = next(self.counter)
        executor = futures.ThreadPoolExecutor(1)
        index_path = self.base / ("%s-%s" %(self.count, index_name))
        idx = index.IndexManager(index_path, executor=executor)
        pkg = getattr(self, pkg)
        pkg.copy(idx.path)
        self.dummypath = idx.path / pkg.name
        return idx

    def test_register_archive(self):
        self.im = self.make_one()
        pkgdata, md5 = self.im.register_archive(self.dummypath)
        assert md5 == '3ac58d89cb7f7b718bc6d0beae85c282'
        assert pkgdata

        idxjson = self.im.data_from_path(self.im.datafile_path)
        assert md5 in idxjson
        assert idxjson[md5] == pkgdata

    def test_write_datafile(self):
        """
        create and write archive data to index.json
        """
        self.im = self.make_one()
        data = self.im.write_datafile(hello='computer')
        assert 'hello' in data
        assert self.im.datafile_path.exists()
        assert 'hello' in self.im.data_from_path(self.im.datafile_path)

    def test_write_datafile_w_existing_datafile(self):
        """
        write data to an existing datafile
        """
        self.im = self.make_one()
        data = self.im.write_datafile(hello='computer')
        assert self.im.datafile_path.exists()

        data = self.im.write_datafile(hello='operator')
        assert data['hello'] == 'operator'
        assert self.im.data_from_path(self.im.datafile_path)['hello'] == 'operator'

    def test_regenerate_index_write_index_html_false(self):
        im = self.make_one()
        im.write_index_html = False
        home, leaves = im.regenerate_all()
        pth = im.path
        assert home is None
        assert not (pth / im.root_index_file).exists()

    def test_regenerate_index(self):
        self.im = self.make_one()
        home, leaves = self.im.regenerate_all()
        pth = self.im.path
        file_structure = [(x.parent.name, x.name) for x in pth.walk()]
        index_name = u'%s-test-index' %self.count
        expected = [(index_name, u'dummypackage'),
                    (u'dummypackage', u'index.html'),
                    (path(u'dummypackage'), path(u'index.json')),
                    (index_name, u'dummypackage-0.0dev.tar.gz'),
                    (index_name, u'index.html')]

        assert len(leaves) == 1
        assert leaves[0].exists()
        assert leaves[0].name == 'index.html'
        assert leaves[0].parent.name == 'dummypackage'

        etxt = pprint(sorted(str(x) for x in expected))
        fstxt = pprint(sorted(str(x) for x in file_structure))
        assert set(expected).issubset(file_structure), \
               textwrap.dedent("""
               Expected not a subset of result::

               expected:

                %s

               actual:

                %s""") %(etxt, fstxt)

    @patch('cheeseprism.index.IndexManager.regenerate_leaf')
    def test_regenerate_leaf_event(self, rl):
        """
        Cover event subscriber
        """
        from cheeseprism.event import PackageAdded
        from cheeseprism.index import rebuild_leaf
        self.im = self.make_one()
        event = PackageAdded(self.im, here / path('dummypackage2/dist/dummypackage-0.1.tar.gz'))
        out = rebuild_leaf(event)
        assert out is not None
        assert rl.call_args == (('dummypackage',), {})

    def test_regenerate_leaf(self):
        self.im = self.make_one()
        [x for x in self.im.regenerate_all()]
        leafindex = self.im.path / 'dummypackage/index.html'
        new_arch = here / path('dummypackage2/dist/dummypackage-0.1.tar.gz')
        new_arch.copy(self.im.path)
        added = self.im.path / new_arch.name

        before_txt = leafindex.text()
        info = self.im.pkginfo_from_file(added)
        out = self.im.regenerate_leaf(info.name)
        assert before_txt != out.text()

    @patch('pyramid.threadlocal.get_current_registry')
    def test_bulk_add_pkg(self, getreg):
        from cheeseprism.index import bulk_add_pkgs
        self.im = self.make_one()
        pkg = stuf(name='dummypackage', version='0.1',
                   filename=self.dummy.name)
        pkgs = pkg,
        index = Mock(name='index')
        index.path = self.im.path
        leaves, archs = bulk_add_pkgs(index, pkgs)

        assert len(archs) == 1
        assert len(leaves) == 1
        assert 'dummypackage' in leaves
        assert archs[0].basename() == u'dummypackage-0.0dev.tar.gz'
        assert index.regenerate_leaf.called

    def tearDown(self):
        logger.debug("teardown: %s", self.count)
        if self.base.exists():
            dirs = self.base.dirs()
            logger.info(pprint(dirs))
            logger.info(pprint([x.rmtree() for x in dirs]))

def test_group_by_magnitude():
    from cheeseprism.index import IndexManager
    fiver = range(5)
    assert IndexManager.group_by_magnitude(fiver)[0] == fiver
    assert next(IndexManager.group_by_magnitude(range(101))) == range(10)
    assert next(IndexManager.group_by_magnitude(range(1001))) == range(100)


class ClassOrStaticMethods(unittest.TestCase):

    def test_move_on_error(self):
        from cheeseprism.index import ArchiveUtil
        exc, path = Mock(), Mock()
        path.basename.return_value = '_path_'
        ArchiveUtil.move_on_error('errors', exc, path)
        assert path.rename.called
        assert path.rename.call_args[0][0] == 'errors/_path_'

    def test_pkginfo_from_file_whl(self):
        """
        .pkginfo_from_file: wheel
        """
        from cheeseprism.index import IndexManager
        with patch('pkginfo.wheel.Wheel', new=Mock(return_value=True)):
            assert IndexManager.pkginfo_from_file('blah.whl') is True

    @patch('pkginfo.bdist.BDist', new=Mock(return_value=True))
    def test_pkginfo_from_file_egg(self):
        """
        .pkginfo_from_file: bdist
        """
        from cheeseprism.index import IndexManager
        assert IndexManager.pkginfo_from_file('blah.egg') is True

    @patch('pkginfo.sdist.SDist', new=Mock(return_value=True))
    def test_pkginfo_from_file_sdist(self):
        """
        .pkginfo_from_file: sdist
        """
        from cheeseprism.index import IndexManager
        for ext in ('.gz','.tgz', '.bz2', '.zip'):
            assert IndexManager.pkginfo_from_file('blah.%s' %ext) is True

    def test_pkginfo_from_bad_ext(self):
        """
        .pkginfo_from_file with unrecognized extension
        """
        from cheeseprism.index import IndexManager
        with self.assertRaises(RuntimeError):
            IndexManager.pkginfo_from_file('adfasdkfha.adkfhalsdk')

    def test_pkginfo_from_bad_ext_handled(self):
        """
        .pkginfo_from_file with unrecognized extension
        """
        from cheeseprism.index import IndexManager
        handler = Mock(name='handler')
        IndexManager.pkginfo_from_file('adfasdkfha.adkfhalsdk', handle_error=handler)
        assert handler.call_args[0][1] == 'adfasdkfha.adkfhalsdk'
        assert isinstance(handler.call_args[0][0], RuntimeError)


    def test_pkginfo_from_no_ext(self):
        """
        .pkginfo_from_file with no extension
        """
        from cheeseprism.index import IndexManager
        with self.assertRaises(RuntimeError):
            IndexManager.pkginfo_from_file('adfasdkfha')


    def test_pkginfo_from_file_exc_and_handler(self):
        """
        .pkginfo_from_file with exception and handler
        """
        from cheeseprism.index import IndexManager
        exc = Exception("BOOM")
        with patch('pkginfo.bdist.BDist', side_effect=exc):
            eh = Mock(name='error_handler')
            IndexManager.pkginfo_from_file('bad.egg', handle_error=eh)
        assert eh.called
        assert eh.call_args[0] == (exc, 'bad.egg'), eh.call_args[0]

    def test_pkginfo_from_file_exc(self):
        """
        .pkginfo_from_file with exception and no handler
        """
        from cheeseprism.index import IndexManager
        exc = ValueError("BOOM")
        with self.assertRaises(ValueError):
            with patch('pkginfo.bdist.BDist', side_effect=exc):
                IndexManager.pkginfo_from_file('bad.egg')

def test_cleanup():
    assert not IndexTestCase.get_base().dirs()

    
def test_noop():
    from cheeseprism.index import noop
    assert noop() is None
    



from .utils import path
import logging
import pkginfo
import re
import time
import traceback


logger = logging.getLogger(__name__)


class ArchiveUtil(object):
    """
    A pickeable object we can pass via mp queues
    """
    EXTS = re.compile(r'^.*(?P<ext>\.egg|\.gz|\.bz2|\.tgz|\.zip|\.whl)$')

    def read(self, (arch, data)):
        pkgdata = None
        if not arch.md5hex in data:
            pkgdata = self.arch_to_add_map(arch)
        return arch.md5hex, pkgdata
    __call__ = read

    def arch_to_add_map(self, arch, error_handler=None):
        pkgi = self.pkginfo_from_file(arch, handle_error=error_handler)
        if pkgi:
            return self.pkginfo_to_pkgdata(arch, pkgi)

    @staticmethod
    def move_on_error(error_folder, exc, path_):
        logger.error(traceback.format_exc())
        path_.rename(path(error_folder) / path_.basename())

    def extension_of(self, path):
        match = self.EXTS.match(str(path))
        if match:
            return match.groupdict()['ext']

    def pkginfo_to_pkgdata(self, arch, pkgi):
        start = time.time()
        return dict(name=pkgi.name,
                    version=pkgi.version,
                    filename=str(arch.name),
                    added=start)

    def pkginfo_from_file(self, path, handle_error=None):
        ext = self.extension_of(path)
        not_recognized = False
        try:
            if ext is not None:
                if ext in set(('.gz','.tgz', '.bz2', '.zip')):
                    return pkginfo.sdist.SDist(path)
                elif ext == '.egg':
                    return pkginfo.bdist.BDist(path)
                elif ext == '.whl':
                    return pkginfo.wheel.Wheel(path)
            not_recognized = True
        except Exception, e:
            if handle_error is not None:
                return handle_error(e, path)
            raise

        if not_recognized is True:
            msg = "Unrecognized extension: %s" %path
            e = RuntimeError("Unrecognized extension: %s" %path)
            if handle_error is not None:
                logger.error(msg)
                return handle_error(e, path)

            raise e



from cheeseprism.index import IndexManager
from .utils import path
from pyramid.security import unauthenticated_userid


def includeme(config):
    for name, func in request_funcs().items():
        reify = not getattr(func, 'skip_reify', False)
        config.add_request_method(func, name=name, reify=reify)


def request_funcs():
    """
    Returns a map of functions to apply to the request

    ATT: No variables can be assigned for this work.
    """
    def userid(request):
        return unauthenticated_userid(request)
    userid.skip_reify = True

    def settings(request):
        return request.registry.settings

    def executor(request):
        return request.registry['cp.executor']

    def index_templates(request):
        return request.registry.settings['cheeseprism.index_templates']

    def file_root(request):
        return path(request.registry.settings['cheeseprism.file_root'])

    def index(request):
        return IndexManager.from_registry(request.registry)

    def index_data_path(request):
        return request.index.datafile_path

    def index_data(request):
        return request.index.data_from_path(request.file_root / request.index_data_path)

    return locals()



import unittest
from test_views import CPDummyRequest


class CPBasicAuthTest(unittest.TestCase):
    def makeone(self, check=None):
        from cheeseprism.auth import BasicAuthenticationPolicy as policy
        if check is None:
            check = policy.noop_check
        return policy(check)

    def setUp(self):
        self.req = CPDummyRequest()
        self.req.environ['HTTP_AUTHORIZATION'] = 'Basic d2hpdDpzZWNyZXQ='
        self.req.headers['Authorization'] = 'Basic d2hpdDpzZWNyZXQ='
        self.req.environ['wsgi.version'] = '1.0'

    def test_authenticated_userid(self):
        policy = self.makeone()
        userid = policy.authenticated_userid(self.req)
        assert userid == 'whit'
        
    def test_authenticated_userid_nocred(self):
        policy = self.makeone()
        del self.req.environ['HTTP_AUTHORIZATION'] 
        userid = policy.authenticated_userid(self.req)
        assert userid is None

    def test_get_cred_good(self):
        from cheeseprism.auth import BasicAuthenticationPolicy as policy
        cred = policy._get_credentials(self.req)
        assert cred == {'login': 'whit', 'password': 'secret'}

    def test_get_cred_bad(self):
        from cheeseprism.auth import BasicAuthenticationPolicy as policy

        self.req.environ['HTTP_AUTHORIZATION'] = 'bleh'
        assert policy._get_credentials(self.req) is None

        self.req.environ['HTTP_AUTHORIZATION'] = 'Basic 123'
        assert policy._get_credentials(self.req) is None

        self.req.environ['HTTP_AUTHORIZATION'] = 'Basic d2hpdCtzZWNyZXQ=\n'
        assert policy._get_credentials(self.req) is None

        self.req.environ['HTTP_AUTHORIZATION'] = 'fah nah'
        assert policy._get_credentials(self.req) is None

    def test_effective_principals(self):
        policy = self.makeone()
        princs = policy.effective_principals(self.req)
        assert princs == ['system.Everyone', 'system.Authenticated', 'whit']

    def test_effective_p_without_cred(self):
        self.req.environ['HTTP_AUTHORIZATION'] = 'Basic d2hpdCtzZWNyZXQ=\n'
        policy = self.makeone()
        assert policy.effective_principals(self.req) == ['system.Everyone']

    def test_effective_p_without_groups(self):
        policy = self.makeone(lambda x, y: None)
        assert policy.effective_principals(self.req) == ['system.Everyone']

    def test_unauth_userid(self):
        policy = self.makeone()
        assert 'whit' == policy.unauthenticated_userid(self.req)
        self.req.environ['HTTP_AUTHORIZATION'] = 'Bad Auth'
        assert policy.unauthenticated_userid(self.req) is None


    def test_remember(self):
        policy = self.makeone()
        assert policy.remember(self.req, 'whit') == []

    def test_forget(self):
        policy = self.makeone()
        assert policy.forget(self.req) == [('WWW-Authenticate', 'Basic realm="Realm"')]
        
        



import jinja2

class EnvFactory(object):
    env_class = jinja2.Environment
    def __init__(self, config):
        self.config = config

    @property
    def loaders(self):
        if self.config:
            loaders = self.config.split(' ')
            for loader in loaders:
                spec = loader.split(':')
                if len(spec) == 1:
                    yield jinja2.FileSystemLoader(spec); continue

                type_, spec = spec
                if type_ == "file":
                    yield jinja2.FileSystemLoader(spec); continue

                if type_ == 'pkg':
                    spec = spec.split('#')
                    if len(spec) == 1: yield jinja2.PackageLoader(spec[0])
                    else: yield jinja2.PackageLoader(*spec)
                    continue
                raise RuntimeError('Loader type not found: %s %s' %(type_, spec))

    @classmethod
    def from_str(cls, config=None):
        factory = cls(config)
        choices = [jinja2.PackageLoader('cheeseprism', 'templates/index')]
        if config: [choices.insert(0, loader) for loader in factory.loaders]
        return factory.env_class(loader=jinja2.ChoiceLoader(choices))



#from time import time
import logging
import xmlrpclib


logger = logging.getLogger(__name__)


class PyPi(object):
    index = 'http://pypi.python.org/pypi'
    pkglist = None
    timeout = 5 * 60

# for autocomplete
##     @classmethod
##     def list_pkgs(cls):
##         now = time()
##         if cls.pkglist is not None:
##             expires, pkglist = cls.pkglist
##             if not expires > now:
##                 logger.debug('found cached list')
##                 return pkglist

##         logger.debug('getting list from pypi')
##         client = xmlrpclib.ServerProxy(cls.index)
##         pkglist = client.list_packages()
##         cls.pkglist = (now + cls.timeout, pkglist)
##         return pkglist

    @classmethod
    def search(cls, package_name):
        client = xmlrpclib.ServerProxy(cls.index)
        return client.package_releases(package_name)

    ### for additional information
    @classmethod
    def package_details(cls, package_name, version_number):
        client = xmlrpclib.ServerProxy(cls.index)
        return client.release_urls(package_name, version_number)

    @classmethod
    def release_urls(cls, package_name, version_number):
        client = xmlrpclib.ServerProxy(cls.index)
        return client.release_urls(package_name, version_number)



import unittest

class TestUpdict(unittest.TestCase):

    def makeone(self):
        from cheeseprism.desc import updict
        self.seed = dict(default=True)

        class Obj(object):
            attr = updict(self.seed)

        self.klass = Obj
        return Obj()

    def setUp(self):
        self.test_obj = self.makeone()
    
    def test_get(self):
        assert 'default' in self.test_obj.attr
        assert self.klass.attr is self.test_obj.attr

    def test_set(self):
        self.test_obj.attr = dict(add=1)
        assert 'add' in self.test_obj.attr

    def test_del(self):
        self.test_obj.attr = dict(add=1)
        assert 'add' in self.test_obj.attr
        del self.test_obj.attr
        assert not 'add' in self.test_obj.attr



from cheeseprism.utils import path

here = path(__name__).parent
testdir = here / 'test-indexes/test-main'

def test_main():
    """
    sanity check for code that creates wsgi app
    """
    from cheeseprism.wsgiapp import main
    globconf = dict()
    app = main(globconf, **{'cheeseprism.index_templates':'',
                            'cheeseprism.file_root': testdir,
                            'cheeseprism.data_json': 'data.json'})
    assert app

def teardown():
    if testdir.exists():
        testdir.rmtree()



from cheeseprism import event
from cheeseprism import pipext
from cheeseprism import resources
from cheeseprism.rpc import PyPi
from .utils import path
from pyramid.httpexceptions import HTTPFound
from pyramid.i18n import TranslationStringFactory
from pyramid.path import DottedNameResolver as dnr
from pyramid.view import view_config
from pyramid_jinja2 import renderer_factory
from urllib2 import HTTPError
from urllib2 import URLError
from webob import exc
import logging
import requests
import tempfile


resolve = dnr(None).maybe_resolve
logger = logging.getLogger(__name__)


_ = TranslationStringFactory('CheesePrism')


@view_config(name='instructions', renderer='instructions.html', context=resources.App)
def instructions(context, request):
    return {'page': 'instructions'}


@view_config(name='index', renderer='index/home.html', context=resources.App)
def index_view(context, request):
    index = request.index
    data = index.index_data.copy()
    items = index.projects_from_archives()
    data['packages'] = [dict(name=key, url=str(path(index.urlbase) / key )) \
                        for key, value in items]
    return data

@view_config(renderer='index.html', context=resources.App)
@view_config(name='simple', context=resources.App)
def upload(context, request):
    """
    The interface for disutils upload
    """
    if request.method == 'POST':
        if not hasattr(request.POST['content'], 'file'):
            raise RuntimeError('No file attached')

        fieldstorage = request.POST['content']
        filename = fieldstorage.filename
        logger.info("%s posted", filename)

        dest = path(request.file_root) / request.namer(filename)
        dest.write_bytes(fieldstorage.file.read())
        try:
            request.registry.notify(event.PackageAdded(request.index, path=dest))
            request.response.headers['X-Swalow-Status'] = 'SUCCESS'
            return request.response
        except :
            logger.exception("Processing of %s failed", filename)
            raise
    return {}


@view_config(name='find-packages', renderer='find_packages.html', context=resources.App)
def find_package(context, request):
    releases = None
    search_term = None
    if request.method == "POST":
        search_term = request.POST['search_box']
        releases = PyPi.search(search_term)
    url = request.resource_url(context, request.view_name)
    return dict(releases=releases, search_term=search_term, here=url)


def from_pypi(request, fpkgs='/find-packages'):
    """
    @@ convert to use action on a post rather than on a get
    """
    name = request.matchdict['name']
    version = request.matchdict['version']
    dists = PyPi.release_urls(name, version)
    flash = request.session.flash
    if not dists:
        flash("%s-%s not found" %(name, version))
        return HTTPFound(fpkgs)

    candidates = [x for x in dists if request.index.SDIST_EXT.match(x['filename'])]

    if candidates[0]['md5_digest'] in request.index_data:
        logger.debug('Package %s-%s already in index' %(name, version))
        return HTTPFound('/index/%s' %name)

    details = candidates[0]
    url = details['url']
    filename = details['filename']
    newfile = None
    try:
        resp = requests.get(url)
        newfile = request.file_root / filename
        newfile.write_bytes(resp.content)
    except HTTPError, e:
        error = "HTTP Error: %d %s - %s" %(e.code, exc.status_map[e.code].title, url)
        logger.error(error)
        flash(error)
    except URLError, e:
        logger.error("URL Error: %s, %s", e.reason , url)
        flash('Url error attempting to grab %s: %s' %(url, e.reason))

    if newfile is not None:
        try:
            added_event = event.PackageAdded(request.index, path=newfile)
            request.registry.notify(added_event)
            flash('%s-%s was installed into the index successfully.' % (name, version))
            return HTTPFound('/index/%s' %name)
        except Exception, e:
            flash('Issue with adding %s to index: See logs: %s' % (newfile.name, e))

    return HTTPFound(fpkgs)
package = from_pypi


@view_config(name='regenerate-index', renderer='regenerate.html', context=resources.App)
def regenerate_index(context, request):
    if request.method == 'POST':
        logger.debug("Regenerate index")
        homefile, leaves = request.index.regenerate_all()
        request.index.update_data()
        return HTTPFound('/index')
    return {'disabled': request.disable_regen}


@view_config(name='load-requirements',
             renderer='requirements_upload.html',
             context=resources.App)
def from_requirements(context, request):
    if request.method == "POST":
        req_text = request.POST['req_file'].file.read()
        filename = path(tempfile.gettempdir()) / 'temp-req.txt'
        filename.write_text(req_text)

        names = []
        pkgdatas = {}
        rd_class = pipext.RequirementDownloader
        requirement_set, finder = rd_class.req_set_from_file(filename, request.file_root)
        downloader = rd_class(requirement_set, finder, seen=set(request.index_data))

        for pkginfo, outfile in downloader.download_all():
            pkgdatas[outfile] = \
                request.index.pkginfo_to_pkgdata(outfile, pkginfo)
            names.append(pkginfo.name)

        request.registry.notify(event.IndexUpdate(
            request.index_data_path, request.index, pkgdatas))

        flash = request.session.flash
        if names:
            flash('The following packages were installed '
                  'from the requirements file: %s' % ", ".join(names))

        if downloader.skip:
            for dl in (x.filename for x in downloader.skip):
                flash("Skipped (already in index): %s" %dl)

        if downloader.errors:
            for error in downloader.errors:
                flash('Download issue: %s' %error)

        return HTTPFound('/load-requirements')
    return {}


def includeme(config):
    config.scan(__name__)
    config.add_route('package', 'package/{name}/{version}')
    config.add_view('.views.from_pypi', route_name='package')
    config.add_static_view('static', 'static')
    config.include('pyramid_jinja2')
    config.add_renderer('.html', renderer_factory)

    namer = resolve(config.registry.settings.get('cheeseprism.namer', 'cheeseprism.utils.secure_filename'))
    assert namer and callable(namer), "Issue with namer: %s" %namer
    config.registry['cheeseprism.namer'] = namer
    config.add_request_method(lambda req: namer, name='namer', reify=True)



from cheeseprism.index import IndexManager
from functools import partial
from .utils import path
from pip.exceptions import DistributionNotFound
from pip.index import PackageFinder
from pip.locations import build_prefix, src_prefix
from pip.req import RequirementSet, parse_requirements
from urllib2 import HTTPError
import logging
import requests
import tarfile
import tempfile
import zipfile


logger = logging.getLogger(__name__)


class RequirementDownloader(object):
    build_dir = path(build_prefix).abspath()
    pkginfo_from_file = IndexManager.pkginfo_from_file
    parse_requirements = staticmethod(parse_requirements)

    def __init__(self, req_set, finder=None, upgrade=False, seen=None):
        #@@ start with req_set??
        self.req_set = req_set
        self.upgrade = False
        download_dir = req_set.download_dir
        self.download_dir = path(download_dir)
        self.finder = finder
        self.seen = seen
        self.skip = []
        if self.seen is None:
            self.seen = set()
        self.errors = []

    # toupe for pip
    options = type('Options', (), dict(skip_requirements_regex='',
                                       default_vcs=''))

    @classmethod
    def req_set_from_file(cls, filename, download_dir, deplinks=None):
        src_dir = path(src_prefix).abspath()

        finder = cls.package_finder(deplinks)

        requirement_set = RequirementSet(
            build_dir=cls.build_dir,
            src_dir=src_dir,
            download_dir=download_dir,
            download_cache=None,
            upgrade=False,
            ignore_installed=True,
            ignore_dependencies=False)

        options = cls.options()
        names = []
        for req in cls.parse_requirements(filename, finder=finder, options=options):
            requirement_set.add_requirement(req)
            names.append(req.req)
        yield requirement_set
        yield finder

    @staticmethod
    def readzip(archive, name):
        return archive.read(name)

    @staticmethod
    def readtar(archive, name):
        return archive.extractfile(name).read()    

    @staticmethod
    def find_file(names, tail):
        try:
            return next(n for n in names if n.endswith(tail))
        except StopIteration:
            return None
        
    @classmethod
    def depinfo_for_file(cls, filename):
        if filename.endswith('.zip'):
            archive = zipfile.ZipFile(filename)
            names = archive.namelist()
            read = partial(cls.readzip, archive)
        elif filename.endswith('gz') or filename.endswith('bz2'):
            archive = tarfile.TarFile.open(filename)
            names = archive.getnames()
            read = partial(cls.readtar, archive)
        dl_file = cls.find_file(names, '.egg-info/dependency_links.txt')
        reqs_file = cls.find_file(names, '.egg-info/requires.txt')        
        deplinks = dl_file and read(dl_file) or ''
        requires = reqs_file and read(reqs_file) or ''
        return [x.strip() for x in deplinks.split('\n') if x],\
               [x.strip() for x in requires.split('\n') if x and not x.startswith('[')]

    def download_url(self, link):
        target_url = link.url.split('#', 1)[0]
        resp = requests.get(target_url)
        logger.info('Downloading: %s' %target_url)
        outfile = self.download_dir / link.filename
        outfile.write_bytes(resp.content)
        # requests.iter_content
        pkginfo = self.pkginfo_from_file(outfile)
        return pkginfo, outfile

    def temp_req(self, name, content=None):
        fp = path(tempfile.gettempdir()) / ('temp-req-%s.txt' %name)
        if content:
            logger.debug("Reqs for %s:\n%s", name, content)
            fp.write_text(content)
        return fp

    def handle_requirement(self, req, finder):
        """
        Download requirement, return a new requirement set of
        requirements dependencies.
        """
        if req.editable:
            msg = "Editables not supported: %s" %req
            logger.warn(msg)
            self.errors.append("%s: %s" %(req, msg)) 
            return

        try:
            url = finder.find_requirement(req, self.upgrade)
        except DistributionNotFound:
            msg = "No distribution found for %s" %req.name
            logger.warn(msg)
            self.errors.append(msg)
            return
        
        if url.hash in self.seen:
            logger.debug('Seen: %s', url)
            self.skip.append(url)
            return 
            
        try:
            pkginfo, outfile = self.download_url(url)
        except HTTPError, e:
            msg = "Issue with download: %s" %e
            logger.error(msg)
            self.errors.append("%s: %s" %(req, msg)) 
            return
        except TypeError:
            raise
        except Exception, e:
            msg = "Issue with archive: %s" %e
            logger.error(msg)
            self.errors.append("%s: %s" %(req, msg)) 
            return

        self.seen.add(outfile.md5hex)
        deplinks, reqs = self.depinfo_for_file(outfile)
        if not reqs:
            return pkginfo, outfile, None
        
        content = "\n".join(reqs)
        pkg = "%s-%s" %(pkginfo.name, pkginfo.version)
        req_set, _ = self.req_set_from_file(self.temp_req(pkg, content),
                                            self.download_dir,
                                            deplinks=deplinks)
        return pkginfo, outfile, req_set,

    pkg_finder_class = PackageFinder
    index_urls = ['http://pypi.python.org/simple']

    def download_all(self, req_set=None, finder=None):
        if req_set is None:
            req_set = self.req_set
        if finder is None:
            finder = self.finder or self.package_finder(None)

        for req in req_set.requirements.values():
            output = self.handle_requirement(req, finder)
            if output is None:
                continue
            self.seen.add(req)
            pkginfo, outfile, req_set = output
            yield pkginfo, outfile
            if req_set is not None:
                logger.info("Dependencies determined: %s" %req_set.requirements.keys())
                for pkginfo, outfile in self.download_all(req_set, finder):
                    yield pkginfo, outfile

    @classmethod
    def package_finder(cls, deplinks, index_urls=None):
        iu = [x for x in cls.index_urls]
        if index_urls:
            iu.extend(index_urls)
        finder = cls.pkg_finder_class(find_links=[], index_urls=iu)
        if deplinks:
            finder.add_dependency_links(deplinks)
        return finder






from paste.httpheaders import AUTHORIZATION
from paste.httpheaders import WWW_AUTHENTICATE
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.security import Authenticated
from pyramid.security import Everyone
from zope.interface import implements
import binascii


class BasicAuthenticationPolicy(object):
    """ A :app:`Pyramid` :term:`authentication policy` which
    obtains data from basic authentication headers.

    Constructor Arguments

    ``check``

        A callback passed the credentials and the request,
        expected to return None if the userid doesn't exist or a sequence
        of group identifiers (possibly empty) if the user does exist.
        Required.

    ``realm``

        Default: ``Realm``.  The Basic Auth realm string.

    """
    implements(IAuthenticationPolicy)

    def __init__(self, check, realm='Realm'):
        self.check = check
        self.realm = realm

    def authenticated_userid(self, request):
        credentials = self._get_credentials(request)
        if credentials is None:
            return None
        userid = credentials['login']
        if self.check(credentials, request) is not None: # is not None!
            return userid

    @staticmethod
    def _get_credentials(request):
        authorization = AUTHORIZATION(request.environ)
        try:
            authmeth, auth = authorization.split(' ', 1)
        except ValueError: # not enough values to unpack
            return None
        if authmeth.lower() == 'basic':
            try:
                auth = auth.strip().decode('base64')
            except binascii.Error: # can't decode
                return None
            try:
                login, password = auth.split(':', 1)
            except ValueError: # not enough values to unpack
                return None
            return {'login':login, 'password':password}

        return None

    def effective_principals(self, request):
        effective_principals = [Everyone]
        credentials = self._get_credentials(request)
        if credentials is None:
            return effective_principals
        userid = credentials['login']
        groups = self.check(credentials, request)
        if groups is None: # is None!
            return effective_principals
        effective_principals.append(Authenticated)
        effective_principals.append(userid)
        effective_principals.extend(groups)
        return effective_principals

    def unauthenticated_userid(self, request):
        creds = self._get_credentials(request)
        if creds is not None:
            return creds['login']
        return None

    def remember(self, request, principal, **kw):
        return []

    def forget(self, request):
        head = WWW_AUTHENTICATE.tuples('Basic realm="%s"' % self.realm)
        return head

    @staticmethod
    def noop_check(credentials, request):
        return []



from setuptools import setup, find_packages
import sys, os

version = '0.0'

setup(name='dummypackage',
      version=version,
      description="",
      long_description="""\
""",
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='',
      author='',
      author_email='',
      url='',
      license='',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          # -*- Extra requirements: -*-
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



import unittest 
from mock import patch


class TestPyPiXMLRPC(unittest.TestCase):
    @patch('xmlrpclib.ServerProxy')
    def test_search(self, sp):
        from cheeseprism.rpc import PyPi
        out = PyPi.search('five.intid')
        assert sp.called
        assert out

    @patch('xmlrpclib.ServerProxy')
    def test_details(self, sp):
        from cheeseprism.rpc import PyPi
        out = PyPi.package_details('wicked', '1.0')
        ((index,)), _ = sp.call_args
        assert index == PyPi.index
        assert sp.called
        assert out



import jinja2.loaders
from cheeseprism.utils import path


def test_envfactory_makes_loaders():
    from cheeseprism.index import EnvFactory
    env = EnvFactory.from_str('pkg:cheeseprism pkg:cheeseprism#templates/fake file:/tmp/ %s' %path(__file__).parent)
    assert isinstance(env.loader, jinja2.loaders.ChoiceLoader)
    assert len(env.loader.loaders) == 5
    # is this backwards?
    assert env.loader.loaders[0].searchpath[0].endswith('tests'),  env.loader.loaders[1].searchpath
    assert env.loader.loaders[1].searchpath == ['/tmp/']
    assert env.loader.loaders[-1].package_path == 'templates/index', env.loader.loaders[-1].package_path
    assert env.loader.loaders[-2].package_path == 'templates'
    assert env.loader.loaders[-3].package_path == 'templates/fake'



from setuptools import find_packages
from setuptools import setup
import os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()

requires = ['Jinja2',
            'Paste',
            'futures',
            'path.py',
            'pkginfo>=1.2b1',
            'pyramid',
            'pyramid_jinja2',
            'requests',
            'pip',
            'more-itertools']

version='0.4.0b2'

setup(name='CheesePrism',
      version=version,
      description='CheesePrism: your personal cheeseshop',
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Framework :: Pylons",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        ],
      author='Whit Morriss (et al.)',
      author_email='whit-at-surveymonkey-dot-com',
      url='https://github.com/whitmo/CheesePrism',
      keywords='web pyramid pylons',
      packages=find_packages(exclude=['tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      entry_points = """\
      [paste.app_factory]
      main = cheeseprism.wsgiapp:main
      [console_scripts]
      cp-mc = cheeseprism.scripts:mc
      """
      )



"""
Classes, subscribers and functions for dealing with index management
"""
from . import event
from .archiveutil import ArchiveUtil
from .desc import template
from .desc import updict
from .jenv import EnvFactory
from .utils import benchmark
from functools import partial
from more_itertools import chunked
from .utils import path
from pyramid import threadlocal
from pyramid.events import ApplicationCreated
from pyramid.events import subscriber
from pyramid.path import DottedNameResolver as dnr
from pyramid.settings import asbool
import json
import logging
import re
import threading
import time


logger = logging.getLogger(__name__)


class IndexManager(object):
    """
    Manages the static file index
    """

    root_index_file = 'index.html'
    EXTS = re.compile(r'^.*(?P<ext>\.egg|\.gz|\.bz2|\.tgz|\.zip)$')
    SDIST_EXT = re.compile(r'^.*(?P<ext>\.gz|\.bz2|\.tgz|\.zip)$')

    leaf_name = 'leaf.html'
    home_name = 'index.html'
    def_index_title = 'CheesePrism'
    leaf_data = updict()
    index_data = updict(title=def_index_title,
                        index_title=def_index_title,
                        description="Welcome to the CheesePrism")
    datafile_name = "index.json"
    index_data_lock = threading.Lock()

    leaf_template = template('leaf.html')
    home_template = template('home.html')

    at = archive_tool = ArchiveUtil()
    _move_on_error = at.move_on_error

    pkginfo_to_pkgdata = at.pkginfo_to_pkgdata
    pkginfo_from_file = at.pkginfo_from_file
    extension_of = at.extension_of


    def __init__(self, index_path, template_env=None,
                 arch_baseurl='/index/', urlbase='', index_data={},
                 leaf_data={}, error_folder='_errors', executor=None,
                 write_index_html=True):

        self.urlbase = urlbase
        self.write_index_html = write_index_html
        self.arch_baseurl = arch_baseurl
        self.template_env = template_env

        if not self.template_env:
            self.template_env = self.default_env_factory('')
        self.index_data = index_data.copy()
        self.leaf_data = leaf_data.copy()
        self.path = path(index_path).makedirs_p()
        self.home_file = self.path / self.root_index_file
        self.datafile_path = self.path / self.datafile_name

        self.error_folder = self.path / error_folder
        if not (self.error_folder.exists() and self.error_folder.isdir()):
            if self.error_folder.endswith('/') and not self.error_folder.isdir():
                self.error_folder.parent.remove_p()
            self.error_folder.makedirs()

        self.move_on_error = partial(self._move_on_error, self.error_folder)
        self.arch_to_add_map = partial(self.at.arch_to_add_map,
                                       error_handler=self.move_on_error)
        self.executor = executor

    @classmethod
    def from_registry(cls, registry):
        settings = registry.settings
        executor = registry['cp.executor']
        env = registry['cp.index_templates']
        return cls.from_settings(settings, executor, env)

    @classmethod
    def from_settings(cls, settings, executor=None, env=None):
        file_root = path(settings['cheeseprism.file_root'])
        if not file_root.exists():
            file_root.makedirs()

        urlbase = settings.get('cheeseprism.urlbase', '')
        write_index_html = asbool(settings.get('cheeseprism.write_index_html', 'true'))
        abu = settings.get('cheeseprism.archive.urlbase', '..')
        idx_tmplt = settings.get('cheeseprism.index_templates', '')
        env = EnvFactory.from_str(idx_tmplt)

        return cls(settings['cheeseprism.file_root'],
                   urlbase=urlbase,
                   arch_baseurl=abu,
                   template_env=env,
                   executor=executor,
                   write_index_html=write_index_html)

    @property
    def default_env_factory(self):
        return EnvFactory.from_str

    @property
    def files(self):
        return (x for x in self.path.files() if self.archive_tool.EXTS.match(x))

    def projects_from_archives(self):
        with benchmark('-- collected projects'):
            projects = {}
            paths = (self.path / item for item in self.files)
            arch_info = partial(pki_ff, handle_error=self.move_on_error)
            results = [info for info in self.executor.map(arch_info, paths)]
            for itempath, info in results:
                projects.setdefault(info.name, []).append((info, itempath))

        with benchmark('-- sorted projects'):
            return sorted(projects.items())

    def regenerate_leaf(self, leafname):
        files = self.path.files('%s-*.*' %leafname)
        
        pki_ff = partial(self.at.pkginfo_from_file, handle_error=self.move_on_error)
        versions = ((pki_ff(self.path / item), item) for item in files)

        return self.write_leaf(self.path / leafname, versions)

    def regenerate_all(self):
        items = self.projects_from_archives()
        if self.write_index_html is False:
            yield None
        else:
            with benchmark('-- wrote index.html'):
                yield self.write_index_home(items)

        with benchmark('-- regenerated index'):
            yield [self.write_leaf(self.path / key, value) for key, value in items]

    def write_index_home(self, items):
        logger.info('Write index home: %s', self.home_file)
        data = self.index_data.copy()
        data['packages'] = [dict(name=key, url=str(path(self.urlbase) / key )) \
                            for key, value in items]
        self.home_file.write_text(self.home_template.render(**data))
        return self.home_file

    def write_leaf(self, leafdir, versions, indexhtml="index.html", indexjson="index.json"):
        if not leafdir.exists():
            leafdir.makedirs()

        leafhome = leafdir / indexhtml
        leafjson = leafdir / indexjson

        versions = list(versions)
        title = "%s:%s" %(self.index_data['title'], leafdir.name)
        tversions = (self.leaf_values(leafdir.name, archive)\
                     for info, archive in versions)

        text = self.leaf_template\
               .render(package_title=leafdir.name,
                       title=title,
                       versions=tversions)

        leafhome.write_text(text)
        with self.index_data_lock: #@@ more granular locks
            with open(leafjson, 'w') as jsonout:
                leafdata = [dict(filename=str(fpath.name),
                                 name=dist.name,
                                 version=dist.version,
                                 mtime=fpath.mtime,
                                 ctime=fpath.ctime,
                                 atime=fpath.ctime,
                                 ) for dist, fpath in versions]
                json.dump(leafdata, jsonout)
        leafhome.utime((time.time(), time.time()))
        return leafhome

    def leaf_values(self, leafname, archive):
        url = str(path(self.arch_baseurl) / archive.name)
        return dict(url=url, name=archive.name)

    @staticmethod
    def data_from_path(datafile):
        datafile = path(datafile)
        if datafile.exists():
            with open(datafile) as stream:
                return json.load(stream)
        else:
            logger.error("No datafile found for %s", datafile)
            datafile.write_text("{}")
        return {}

    def _write_datafile(self, **data):
        if self.datafile_path.exists():
            newdata = data
            with open(self.datafile_path) as root:
                    data = json.load(root)
                    data.update(newdata)

        with open(self.datafile_path, 'w') as root:
            json.dump(data, root)
        return data

    def write_datafile(self, with_lock=True, **data):
        if with_lock is True:
            with self.index_data_lock:
                return self._write_datafile(**data)
        return self._write_datafile(**data)

    def reg_data(self, arch):
        pkgdata = self.arch_to_add_map(arch)
        return arch.md5hex, pkgdata,

    def register_archive(self, arch, registry=None):
        """
        Adds an archive to the master data store (index.json)
        """
        md5, pkgdata = self.reg_data(arch)
        self.write_datafile(**{md5:pkgdata})
        return pkgdata, md5

    @staticmethod
    def group_by_magnitude(collection):
        alen = len(collection)
        if alen > 1000:
            return chunked(collection, 100)
        if alen > 100:
            return chunked(collection, 10)
        return [collection]

    def update_data(self, datafile=None, pkgdatas=None):
        if datafile is None:
            datafile = self.datafile_path

        archs = self.files if pkgdatas is None else pkgdatas.keys()
        new = []

        archs_g = self.group_by_magnitude([x for x in archs])
        with benchmark("Rebuilt /index.json"):
            for archs in archs_g:
                with self.index_data_lock:
                    new.extend(self._update_data(archs, datafile))

        pkgs = len(set(x['name'] for x in new))
        logger.info("Inspected %s versions for %s packages" %(len(new), pkgs))
        return new

    def _update_data(self, archs, datafile):
        data = self.data_from_path(datafile)
        new = []
        exe = self.executor

        read = self.archive_tool
        archdata = [(arch, data) for arch in archs]
        for md5, pkgdata in exe.map(read, archdata):
            if pkgdata is not None:
                data[md5] = pkgdata
                new.append(pkgdata)

        self.write_datafile(with_lock=False, **data)
        return new


def pki_ff(path, handle_error=None, func=IndexManager.at.pkginfo_from_file):
    return path, func(path, handle_error=handle_error)


@subscriber(event.IPackageAdded)
def rebuild_leaf(event):
    reg = threadlocal.get_current_registry()
    logger.debug("Adding %s" %(event.path))
    event.im.register_archive(event.path, registry=reg)

    with benchmark("Rebuilt leaf for %s" %(event.name)):
        out = event.im.regenerate_leaf(event.name)
    return out


@subscriber(event.IIndexUpdate)
def bulk_update_index(event):
    new_pkgs = event.index.update_data(event.datafile, pkgdatas=event.pkgdatas)
    return bulk_add_pkgs(event.index, new_pkgs, register=False)


def bulk_add_pkgs(index, new_pkgs):
    """
    Sidestep the event system for efficiency

    @@ requires a prior update
    """
    with benchmark('') as bm:
        leaves = set()
        archs = []

        for data in new_pkgs:
            leaves.add(data['name'])
            archs.append(index.path / data['filename'])

        bm.name = "Bulk add >> register %s archives and rebuild %s leaves"\
           %(len(archs), len(leaves))

        for leaf in leaves:
            try:
                index.regenerate_leaf(leaf)
            except Exception:
                logger.exception('Issue building leaf for %s', leaf)

    return leaves, archs


def bulk_update_index_at_start(event):
    logger.info("--> Checking and updating index")
    reg = event.app.registry

    index = IndexManager.from_registry(reg)
    logger.info("-- %s pkg in %s", len([x for x in index.files]), index.path.abspath())

    new_pkgs = index.update_data()
    leaves, archs = bulk_add_pkgs(index, new_pkgs)

    home_file = index.path / index.root_index_file
    if index.write_index_html is True and (not home_file.exists() or len(leaves)):
        items = index.projects_from_archives()
        index.write_index_home(items)
    return leaves, archs


def async_bulk_update_at_start(event):
    from threading import Thread
    logger.info("Spawning thread to handle bulk update on start")
    Thread(target=bulk_update_index_at_start,
           args=(event,),
           name='bulk-update-on-start').start()


resolve = dnr(None).maybe_resolve
preup_key = 'cheeseprism.preupdate'

def noop(*args, **kw):
    return

def includeme(config):
    config.scan(__name__)
    preup = resolve(config.registry.settings.get(preup_key, 'cheeseprism.index.noop'))
    if preup:
        config.registry[preup_key] = preup
    if asbool(config.registry.settings.get('cheeseprism.async_restart', False)):
        config.add_subscriber(async_bulk_update_at_start, ApplicationCreated)
    else:
        config.add_subscriber(bulk_update_index_at_start, ApplicationCreated)
